(function(){
'use strict'

var _$normalize_1 = normalize;

function normalize (arr, dim) {
	if (!arr || arr.length == null) { throw Error('Argument should be an array') }

	if (dim == null) { dim = 1 }
	else { dim = Math.floor(dim) }

	var bounds = Array(dim * 2)

	for (var offset = 0; offset < dim; offset++) {
		var max = -Infinity, min = Infinity, i = offset, l = arr.length;

		for (; i < l; i+=dim) {
			if (arr[i] > max) { max = arr[i]; }
			if (arr[i] < min) { min = arr[i]; }
		}

		bounds[offset] = min
		bounds[dim + offset] = max
	}

	return bounds;
}
"use strict"

function compileSearch(funcName, predicate, reversed, extraArgs, earlyOut) {
  var code = [
    "function ", funcName, "(a,l,h,", extraArgs.join(","),  "){",
earlyOut ? "" : "var i=", (reversed ? "l-1" : "h+1"),
";while(l<=h){\
var m=(l+h)>>>1,x=a[m]"]
  if(earlyOut) {
    if(predicate.indexOf("c") < 0) {
      code.push(";if(x===y){return m}else if(x<=y){")
    } else {
      code.push(";var p=c(x,y);if(p===0){return m}else if(p<=0){")
    }
  } else {
    code.push(";if(", predicate, "){i=m;")
  }
  if(reversed) {
    code.push("l=m+1}else{h=m-1}")
  } else {
    code.push("h=m-1}else{l=m+1}")
  }
  code.push("}")
  if(earlyOut) {
    code.push("return -1};")
  } else {
    code.push("return i};")
  }
  return code.join("")
}

function compileBoundsSearch(predicate, reversed, suffix, earlyOut) {
  var result = new Function([
  compileSearch("A", "x" + predicate + "y", reversed, ["y"], earlyOut),
  compileSearch("P", "c(x,y)" + predicate + "0", reversed, ["y", "c"], earlyOut),
"function dispatchBsearch", suffix, "(a,y,c,l,h){\
if(typeof(c)==='function'){\
return P(a,(l===void 0)?0:l|0,(h===void 0)?a.length-1:h|0,y,c)\
}else{\
return A(a,(c===void 0)?0:c|0,(l===void 0)?a.length-1:l|0,y)\
}}\
return dispatchBsearch", suffix].join(""))
  return result()
}

var _$module_2 = {
  ge: compileBoundsSearch(">=", false, "GE"),
  gt: compileBoundsSearch(">", false, "GT"),
  lt: compileBoundsSearch("<", true, "LT"),
  le: compileBoundsSearch("<=", true, "LE"),
  eq: compileBoundsSearch("-", true, "EQ", true)
}
var _$clamp_3 = clamp

function clamp(value, min, max) {
  return min < max
    ? (value < min ? min : value > max ? max : value)
    : (value < max ? max : value > min ? min : value)
}
var _$module_4 = {};
'use strict'

/* removed: var _$clamp_3 = require('clamp') */;

_$module_4 = toNumber
_$module_4.to = toNumber
_$module_4.from = fromNumber

function toNumber (rgba, normalized) {
	if(normalized == null) { normalized = true }

	var r = rgba[0], g = rgba[1], b = rgba[2], a = rgba[3]

	if (a == null) { a = normalized ? 1 : 255 }

	if (normalized) {
		r *= 255
		g *= 255
		b *= 255
		a *= 255
	}

	r = _$clamp_3(r, 0, 255) & 0xFF
	g = _$clamp_3(g, 0, 255) & 0xFF
	b = _$clamp_3(b, 0, 255) & 0xFF
	a = _$clamp_3(a, 0, 255) & 0xFF

	//hi-order shift converts to -1, so we can't use <<24
	var n = (r * 0x01000000) + (g << 16) + (b << 8) + (a)

	return n
}

function fromNumber (n, normalized) {
	n = +n

	var r = n >>> 24
	var g = (n & 0x00ff0000) >>> 16
	var b = (n & 0x0000ff00) >>> 8
	var a = n & 0x000000ff

	if (normalized === false) { return [r, g, b, a] }

	return [r/255, g/255, b/255, a/255]
}
var _$module_5 = {
	"aliceblue": [240, 248, 255],
	"antiquewhite": [250, 235, 215],
	"aqua": [0, 255, 255],
	"aquamarine": [127, 255, 212],
	"azure": [240, 255, 255],
	"beige": [245, 245, 220],
	"bisque": [255, 228, 196],
	"black": [0, 0, 0],
	"blanchedalmond": [255, 235, 205],
	"blue": [0, 0, 255],
	"blueviolet": [138, 43, 226],
	"brown": [165, 42, 42],
	"burlywood": [222, 184, 135],
	"cadetblue": [95, 158, 160],
	"chartreuse": [127, 255, 0],
	"chocolate": [210, 105, 30],
	"coral": [255, 127, 80],
	"cornflowerblue": [100, 149, 237],
	"cornsilk": [255, 248, 220],
	"crimson": [220, 20, 60],
	"cyan": [0, 255, 255],
	"darkblue": [0, 0, 139],
	"darkcyan": [0, 139, 139],
	"darkgoldenrod": [184, 134, 11],
	"darkgray": [169, 169, 169],
	"darkgreen": [0, 100, 0],
	"darkgrey": [169, 169, 169],
	"darkkhaki": [189, 183, 107],
	"darkmagenta": [139, 0, 139],
	"darkolivegreen": [85, 107, 47],
	"darkorange": [255, 140, 0],
	"darkorchid": [153, 50, 204],
	"darkred": [139, 0, 0],
	"darksalmon": [233, 150, 122],
	"darkseagreen": [143, 188, 143],
	"darkslateblue": [72, 61, 139],
	"darkslategray": [47, 79, 79],
	"darkslategrey": [47, 79, 79],
	"darkturquoise": [0, 206, 209],
	"darkviolet": [148, 0, 211],
	"deeppink": [255, 20, 147],
	"deepskyblue": [0, 191, 255],
	"dimgray": [105, 105, 105],
	"dimgrey": [105, 105, 105],
	"dodgerblue": [30, 144, 255],
	"firebrick": [178, 34, 34],
	"floralwhite": [255, 250, 240],
	"forestgreen": [34, 139, 34],
	"fuchsia": [255, 0, 255],
	"gainsboro": [220, 220, 220],
	"ghostwhite": [248, 248, 255],
	"gold": [255, 215, 0],
	"goldenrod": [218, 165, 32],
	"gray": [128, 128, 128],
	"green": [0, 128, 0],
	"greenyellow": [173, 255, 47],
	"grey": [128, 128, 128],
	"honeydew": [240, 255, 240],
	"hotpink": [255, 105, 180],
	"indianred": [205, 92, 92],
	"indigo": [75, 0, 130],
	"ivory": [255, 255, 240],
	"khaki": [240, 230, 140],
	"lavender": [230, 230, 250],
	"lavenderblush": [255, 240, 245],
	"lawngreen": [124, 252, 0],
	"lemonchiffon": [255, 250, 205],
	"lightblue": [173, 216, 230],
	"lightcoral": [240, 128, 128],
	"lightcyan": [224, 255, 255],
	"lightgoldenrodyellow": [250, 250, 210],
	"lightgray": [211, 211, 211],
	"lightgreen": [144, 238, 144],
	"lightgrey": [211, 211, 211],
	"lightpink": [255, 182, 193],
	"lightsalmon": [255, 160, 122],
	"lightseagreen": [32, 178, 170],
	"lightskyblue": [135, 206, 250],
	"lightslategray": [119, 136, 153],
	"lightslategrey": [119, 136, 153],
	"lightsteelblue": [176, 196, 222],
	"lightyellow": [255, 255, 224],
	"lime": [0, 255, 0],
	"limegreen": [50, 205, 50],
	"linen": [250, 240, 230],
	"magenta": [255, 0, 255],
	"maroon": [128, 0, 0],
	"mediumaquamarine": [102, 205, 170],
	"mediumblue": [0, 0, 205],
	"mediumorchid": [186, 85, 211],
	"mediumpurple": [147, 112, 219],
	"mediumseagreen": [60, 179, 113],
	"mediumslateblue": [123, 104, 238],
	"mediumspringgreen": [0, 250, 154],
	"mediumturquoise": [72, 209, 204],
	"mediumvioletred": [199, 21, 133],
	"midnightblue": [25, 25, 112],
	"mintcream": [245, 255, 250],
	"mistyrose": [255, 228, 225],
	"moccasin": [255, 228, 181],
	"navajowhite": [255, 222, 173],
	"navy": [0, 0, 128],
	"oldlace": [253, 245, 230],
	"olive": [128, 128, 0],
	"olivedrab": [107, 142, 35],
	"orange": [255, 165, 0],
	"orangered": [255, 69, 0],
	"orchid": [218, 112, 214],
	"palegoldenrod": [238, 232, 170],
	"palegreen": [152, 251, 152],
	"paleturquoise": [175, 238, 238],
	"palevioletred": [219, 112, 147],
	"papayawhip": [255, 239, 213],
	"peachpuff": [255, 218, 185],
	"peru": [205, 133, 63],
	"pink": [255, 192, 203],
	"plum": [221, 160, 221],
	"powderblue": [176, 224, 230],
	"purple": [128, 0, 128],
	"rebeccapurple": [102, 51, 153],
	"red": [255, 0, 0],
	"rosybrown": [188, 143, 143],
	"royalblue": [65, 105, 225],
	"saddlebrown": [139, 69, 19],
	"salmon": [250, 128, 114],
	"sandybrown": [244, 164, 96],
	"seagreen": [46, 139, 87],
	"seashell": [255, 245, 238],
	"sienna": [160, 82, 45],
	"silver": [192, 192, 192],
	"skyblue": [135, 206, 235],
	"slateblue": [106, 90, 205],
	"slategray": [112, 128, 144],
	"slategrey": [112, 128, 144],
	"snow": [255, 250, 250],
	"springgreen": [0, 255, 127],
	"steelblue": [70, 130, 180],
	"tan": [210, 180, 140],
	"teal": [0, 128, 128],
	"thistle": [216, 191, 216],
	"tomato": [255, 99, 71],
	"turquoise": [64, 224, 208],
	"violet": [238, 130, 238],
	"wheat": [245, 222, 179],
	"white": [255, 255, 255],
	"whitesmoke": [245, 245, 245],
	"yellow": [255, 255, 0],
	"yellowgreen": [154, 205, 50]
};
var _$module_11 = function () {
    var arguments$1 = arguments;

    for (var i = 0; i < arguments.length; i++) {
        if (arguments$1[i] !== undefined) { return arguments$1[i]; }
    }
};
'use strict';
var toString = Object.prototype.toString;

var _$module_15 = function (x) {
	var prototype;
	return toString.call(x) === '[object Object]' && (prototype = Object.getPrototypeOf(x), prototype === null || prototype === Object.getPrototypeOf({}));
};
var _$parse_7 = {};
(function (global){
/**
 * @module color-parse
 */

'use strict'

var names = _$module_5
var isObject = _$module_15
var defined = _$module_11

_$parse_7 = parse

/**
 * Base hues
 * http://dev.w3.org/csswg/css-color/#typedef-named-hue
 */
//FIXME: use external hue detector
var baseHues = {
	red: 0,
	orange: 60,
	yellow: 120,
	green: 180,
	blue: 240,
	purple: 300
}

/**
 * Parse color from the string passed
 *
 * @return {Object} A space indicator `space`, an array `values` and `alpha`
 */
function parse (cstr) {
	var m, parts = [], alpha = 1, space

	if (typeof cstr === 'string') {
		//keyword
		if (names[cstr]) {
			parts = names[cstr].slice()
			space = 'rgb'
		}

		//reserved words
		else if (cstr === 'transparent') {
			alpha = 0
			space = 'rgb'
			parts = [0,0,0]
		}

		//hex
		else if (/^#[A-Fa-f0-9]+$/.test(cstr)) {
			var base = cstr.slice(1)
			var size = base.length
			var isShort = size <= 4
			alpha = 1

			if (isShort) {
				parts = [
					parseInt(base[0] + base[0], 16),
					parseInt(base[1] + base[1], 16),
					parseInt(base[2] + base[2], 16)
				]
				if (size === 4) {
					alpha = parseInt(base[3] + base[3], 16) / 255
				}
			}
			else {
				parts = [
					parseInt(base[0] + base[1], 16),
					parseInt(base[2] + base[3], 16),
					parseInt(base[4] + base[5], 16)
				]
				if (size === 8) {
					alpha = parseInt(base[6] + base[7], 16) / 255
				}
			}

			if (!parts[0]) { parts[0] = 0 }
			if (!parts[1]) { parts[1] = 0 }
			if (!parts[2]) { parts[2] = 0 }

			space = 'rgb'
		}

		//color space
		else if (m = /^((?:rgb|hs[lvb]|hwb|cmyk?|xy[zy]|gray|lab|lchu?v?|[ly]uv|lms)a?)\s*\(([^\)]*)\)/.exec(cstr)) {
			var name = m[1]
			var base = name.replace(/a$/, '')
			space = base
			var size = base === 'cmyk' ? 4 : base === 'gray' ? 1 : 3
			parts = m[2].trim()
				.split(/\s*,\s*/)
				.map(function (x, i) {
					//<percentage>
					if (/%$/.test(x)) {
						//alpha
						if (i === size)	{ return parseFloat(x) / 100 }
						//rgb
						if (base === 'rgb') { return parseFloat(x) * 255 / 100 }
						return parseFloat(x)
					}
					//hue
					else if (base[i] === 'h') {
						//<deg>
						if (/deg$/.test(x)) {
							return parseFloat(x)
						}
						//<base-hue>
						else if (baseHues[x] !== undefined) {
							return baseHues[x]
						}
					}
					return parseFloat(x)
				})

			if (name === base) { parts.push(1) }
			alpha = parts[size] === undefined ? 1 : parts[size]
			parts = parts.slice(0, size)
		}

		//named channels case
		else if (cstr.length > 10 && /[0-9](?:\s|\/)/.test(cstr)) {
			parts = cstr.match(/([0-9]+)/g).map(function (value) {
				return parseFloat(value)
			})

			space = cstr.match(/([a-z])/ig).join('').toLowerCase()
		}
	}

	//numeric case
	else if (typeof cstr === 'number') {
		space = 'rgb'
		parts = [cstr >>> 16, (cstr & 0x00ff00) >>> 8, cstr & 0x0000ff]
	}

	//object case - detects css cases of rgb and hsl
	else if (isObject(cstr)) {
		var r = defined(cstr.r, cstr.red, cstr.R, null)

		if (r !== null) {
			space = 'rgb'
			parts = [
				r,
				defined(cstr.g, cstr.green, cstr.G),
				defined(cstr.b, cstr.blue, cstr.B)
			]
		}
		else {
			space = 'hsl'
			parts = [
				defined(cstr.h, cstr.hue, cstr.H),
				defined(cstr.s, cstr.saturation, cstr.S),
				defined(cstr.l, cstr.lightness, cstr.L, cstr.b, cstr.brightness)
			]
		}

		alpha = defined(cstr.a, cstr.alpha, cstr.opacity, 1)

		if (cstr.opacity != null) { alpha /= 100 }
	}

	//array
	else if (Array.isArray(cstr) || global.ArrayBuffer && ArrayBuffer.isView && ArrayBuffer.isView(cstr)) {
		parts = [cstr[0], cstr[1], cstr[2]]
		space = 'rgb'
		alpha = cstr.length === 4 ? cstr[3] : 1
	}

	return {
		space: space,
		values: parts,
		alpha: alpha
	}
}
}).call(this,typeof global !== "undefined" ? global : typeof self !== "undefined" ? self : typeof window !== "undefined" ? window : {})
'use strict'

var _$module_10 = {
	name: 'rgb',
	min: [0,0,0],
	max: [255,255,255],
	channel: ['red', 'green', 'blue'],
	alias: ['RGB']
};
'use strict'

/* removed: var _$module_10 = require('./rgb'); */;

var _$module_9 = {
	name: 'hsl',
	min: [0,0,0],
	max: [360,100,100],
	channel: ['hue', 'saturation', 'lightness'],
	alias: ['HSL'],

	rgb: function(hsl) {
		var h = hsl[0] / 360,
				s = hsl[1] / 100,
				l = hsl[2] / 100,
				t1, t2, t3, rgb, val;

		if (s === 0) {
			val = l * 255;
			return [val, val, val];
		}

		if (l < 0.5) {
			t2 = l * (1 + s);
		}
		else {
			t2 = l + s - l * s;
		}
		t1 = 2 * l - t2;

		rgb = [0, 0, 0];
		for (var i = 0; i < 3; i++) {
			t3 = h + 1 / 3 * - (i - 1);
			if (t3 < 0) {
				t3++;
			}
			else if (t3 > 1) {
				t3--;
			}

			if (6 * t3 < 1) {
				val = t1 + (t2 - t1) * 6 * t3;
			}
			else if (2 * t3 < 1) {
				val = t2;
			}
			else if (3 * t3 < 2) {
				val = t1 + (t2 - t1) * (2 / 3 - t3) * 6;
			}
			else {
				val = t1;
			}

			rgb[i] = val * 255;
		}

		return rgb;
	}
};


//extend rgb
_$module_10.hsl = function(rgb) {
	var r = rgb[0]/255,
			g = rgb[1]/255,
			b = rgb[2]/255,
			min = Math.min(r, g, b),
			max = Math.max(r, g, b),
			delta = max - min,
			h, s, l;

	if (max === min) {
		h = 0;
	}
	else if (r === max) {
		h = (g - b) / delta;
	}
	else if (g === max) {
		h = 2 + (b - r) / delta;
	}
	else if (b === max) {
		h = 4 + (r - g)/ delta;
	}

	h = Math.min(h * 60, 360);

	if (h < 0) {
		h += 360;
	}

	l = (min + max) / 2;

	if (max === min) {
		s = 0;
	}
	else if (l <= 0.5) {
		s = delta / (max + min);
	}
	else {
		s = delta / (2 - max - min);
	}

	return [h, s * 100, l * 100];
};
'use strict'

/* removed: var _$parse_7 = require('color-parse') */;
/* removed: var _$module_9 = require('color-space/hsl') */;
/* removed: var _$clamp_3 = require('clamp') */;

var _$rgba_8 = function rgba (color) {
	var values, i, l

	if (typeof color !== 'string') { throw Error('Argument should be a string') }

	//attempt to parse non-array arguments
	var parsed = _$parse_7(color)

	if (!parsed.space) { return [] }

	values = Array(3)
	values[0] = _$clamp_3(parsed.values[0], 0, 255)
	values[1] = _$clamp_3(parsed.values[1], 0, 255)
	values[2] = _$clamp_3(parsed.values[2], 0, 255)

	if (parsed.space[0] === 'h') {
		values = _$module_9.rgb(values)
	}

	values.push(_$clamp_3(parsed.alpha, 0, 1))

	return values
}
var _$module_12 = function(dtype) {
  switch (dtype) {
    case 'int8':
      return Int8Array
    case 'int16':
      return Int16Array
    case 'int32':
      return Int32Array
    case 'uint8':
      return Uint8Array
    case 'uint16':
      return Uint16Array
    case 'uint32':
      return Uint32Array
    case 'float32':
      return Float32Array
    case 'float64':
      return Float64Array
    case 'array':
      return Array
    case 'uint8_clamped':
      return Uint8ClampedArray
  }
}
'use strict'

/* removed: var _$rgba_8 = require('color-rgba') */;
/* removed: var _$clamp_3 = require('clamp') */;
/* removed: var _$module_12 = require('dtype') */;

var _$normalize_6 = function normalize (color, type) {
	if (type === 'float' || !type) { type = 'array' }
	if (type === 'uint') { type = 'uint8' }
	if (type === 'uint_clamped') { type = 'uint8_clamped' }

	var Ctor = _$module_12(type)
	var output = new Ctor(4)

	//same class does not change values
	if (color instanceof Ctor) {
		output.set(color)
		return
	}

	var normalize = type !== 'uint8' && type !== 'uint8_clamped'

	//consider uint8 array as 0..255 channel values
	if (color instanceof Uint8Array || color instanceof Uint8ClampedArray) {
		output[0] = color[0]
		output[1] = color[1]
		output[2] = color[2]
		output[3] = color[3] != null ? color[3] : 255

		if (normalize) {
			output[0] /= 255
			output[1] /= 255
			output[2] /= 255
			output[3] /= 255
		}

		return output
	}

	//attempt to parse non-array arguments
	if (!color.length || typeof color === 'string') {
		color = _$rgba_8(color)
		color[0] /= 255
		color[1] /= 255
		color[2] /= 255
	}

	//consider every other array type as 0..1 float values
	if (!normalize) {
		output[0] = _$clamp_3(Math.round(color[0] * 255), 0, 255)
		output[1] = _$clamp_3(Math.round(color[1] * 255), 0, 255)
		output[2] = _$clamp_3(Math.round(color[2] * 255), 0, 255)
		output[3] = color[3] == null ? 255 : _$clamp_3(Math.floor(color[3] * 255), 0, 255)
	} else {
		output[0] = color[0]
		output[1] = color[1]
		output[2] = color[2]
		output[3] = color[3] != null ? color[3] : 1
	}

	return output
}
/* removed: var _$module_12 = require('dtype') */;
var _$flattenVertexData_13 = flattenVertexData
function flattenVertexData (data, output, offset) {
  if (!data) { throw new TypeError('must specify data as first parameter') }
  offset = +(offset || 0) | 0

  if (Array.isArray(data) && Array.isArray(data[0])) {
    var dim = data[0].length
    var length = data.length * dim

    // no output specified, create a new typed array
    if (!output || typeof output === 'string') {
      output = new (_$module_12(output || 'float32'))(length + offset)
    }

    var dstLength = output.length - offset
    if (length !== dstLength) {
      throw new Error('source length ' + length + ' (' + dim + 'x' + data.length + ')' +
        ' does not match destination length ' + dstLength)
    }

    for (var i = 0, k = offset; i < data.length; i++) {
      for (var j = 0; j < dim; j++) {
        output[k++] = data[i][j]
      }
    }
  } else {
    if (!output || typeof output === 'string') {
      // no output, create a new one
      var Ctor = _$module_12(output || 'float32')
      if (offset === 0) {
        output = new Ctor(data)
      } else {
        output = new Ctor(data.length + offset)
        output.set(data, offset)
      }
    } else {
      // store output in existing array
      output.set(data, offset)
    }
  }

  return output
}
var _$module_14 = function(strings) {
  if (typeof strings === 'string') { strings = [strings] }
  var exprs = [].slice.call(arguments,1)
  var parts = []
  for (var i = 0; i < strings.length-1; i++) {
    parts.push(strings[i], exprs[i] || '')
  }
  parts.push(strings[i])
  return parts.join('')
}
'use strict';
/* eslint-disable no-unused-vars */
var getOwnPropertySymbols = Object.getOwnPropertySymbols;
var hasOwnProperty = Object.prototype.hasOwnProperty;
var propIsEnumerable = Object.prototype.propertyIsEnumerable;

function toObject(val) {
	if (val === null || val === undefined) {
		throw new TypeError('Object.assign cannot be called with null or undefined');
	}

	return Object(val);
}

function shouldUseNative() {
	try {
		if (!Object.assign) {
			return false;
		}

		// Detect buggy property enumeration order in older V8 versions.

		// https://bugs.chromium.org/p/v8/issues/detail?id=4118
		var test1 = new String('abc');  // eslint-disable-line no-new-wrappers
		test1[5] = 'de';
		if (Object.getOwnPropertyNames(test1)[0] === '5') {
			return false;
		}

		// https://bugs.chromium.org/p/v8/issues/detail?id=3056
		var test2 = {};
		for (var i = 0; i < 10; i++) {
			test2['_' + String.fromCharCode(i)] = i;
		}
		var order2 = Object.getOwnPropertyNames(test2).map(function (n) {
			return test2[n];
		});
		if (order2.join('') !== '0123456789') {
			return false;
		}

		// https://bugs.chromium.org/p/v8/issues/detail?id=3056
		var test3 = {};
		'abcdefghijklmnopqrst'.split('').forEach(function (letter) {
			test3[letter] = letter;
		});
		if (Object.keys(Object.assign({}, test3)).join('') !==
				'abcdefghijklmnopqrst') {
			return false;
		}

		return true;
	} catch (err) {
		// We don't expect any of the above to throw, but better to be safe.
		return false;
	}
}

var _$module_16 = shouldUseNative() ? Object.assign : function (target, source) {
	var arguments$1 = arguments;

	var from;
	var to = toObject(target);
	var symbols;

	for (var s = 1; s < arguments.length; s++) {
		from = Object(arguments$1[s]);

		for (var key in from) {
			if (hasOwnProperty.call(from, key)) {
				to[key] = from[key];
			}
		}

		if (getOwnPropertySymbols) {
			symbols = getOwnPropertySymbols(from);
			for (var i = 0; i < symbols.length; i++) {
				if (propIsEnumerable.call(from, symbols[i])) {
					to[symbols[i]] = from[symbols[i]];
				}
			}
		}
	}

	return to;
};
'use strict'

var CACHE = {}

var _$pick_17 = function pick (src, props) {
	var result = {}

	for (var prop in props) {
		var aliases = props[prop]

		if (CACHE[aliases]) { aliases = CACHE[aliases] }
		else if (typeof aliases === 'string') {
			aliases = CACHE[aliases] = aliases.split(/\s*,\s*|\s+/)
		}

		if (Array.isArray(aliases)) {
			for (var i = 0; i < aliases.length; i++) {
				var alias = aliases[i]

				if (alias in src) {
					result[prop] = src[alias]
					break
				}
			}
		}
		else {
			result[prop] = src[prop]
		}
	}

	return result
}
'use strict'

var _$sortLevels_18 = sortLevels

var INSERT_SORT_CUTOFF = 32

function sortLevels(data_levels, data_points, data_ids, data_weights, n0) {
  if (n0 <= 4*INSERT_SORT_CUTOFF) {
    insertionSort(0, n0 - 1, data_levels, data_points, data_ids, data_weights)
  } else {
    quickSort(0, n0 - 1, data_levels, data_points, data_ids, data_weights)
  }
}

function insertionSort(left, right, data_levels, data_points, data_ids, data_weights) {
  for(var i=left+1; i<=right; ++i) {
    var a_level  = data_levels[i]
    var a_x      = data_points[2*i]
    var a_y      = data_points[2*i+1]
    var a_id     = data_ids[i]
    var a_weight = data_weights[i]

    var j = i
    while(j > left) {
      var b_level = data_levels[j-1]
      var b_x     = data_points[2*(j-1)]
      if(((b_level - a_level) || (a_x - b_x)) >= 0) {
        break
      }
      data_levels[j]      = b_level
      data_points[2*j]    = b_x
      data_points[2*j+1]  = data_points[2*j-1]
      data_ids[j]         = data_ids[j-1]
      data_weights[j]     = data_weights[j-1]
      j -= 1
    }

    data_levels[j]     = a_level
    data_points[2*j]   = a_x
    data_points[2*j+1] = a_y
    data_ids[j]        = a_id
    data_weights[j]    = a_weight
  }
}

function swap(i, j, data_levels, data_points, data_ids, data_weights) {
  var a_level = data_levels[i]
  var a_x     = data_points[2*i]
  var a_y     = data_points[2*i+1]
  var a_id    = data_ids[i]
  var a_weight = data_weights[i]

  data_levels[i]     = data_levels[j]
  data_points[2*i]   = data_points[2*j]
  data_points[2*i+1] = data_points[2*j+1]
  data_ids[i]        = data_ids[j]
  data_weights[i]    = data_weights[j]

  data_levels[j]     = a_level
  data_points[2*j]   = a_x
  data_points[2*j+1] = a_y
  data_ids[j]        = a_id
  data_weights[j]    = a_weight
}

function move(i, j, data_levels, data_points, data_ids, data_weights) {
  data_levels[i]     = data_levels[j]
  data_points[2*i]   = data_points[2*j]
  data_points[2*i+1] = data_points[2*j+1]
  data_ids[i]        = data_ids[j]
  data_weights[i]    = data_weights[j]
}

function rotate(i, j, k, data_levels, data_points, data_ids, data_weights) {
  var a_level = data_levels[i]
  var a_x     = data_points[2*i]
  var a_y     = data_points[2*i+1]
  var a_id    = data_ids[i]
  var a_weight = data_weights[i]

  data_levels[i]     = data_levels[j]
  data_points[2*i]   = data_points[2*j]
  data_points[2*i+1] = data_points[2*j+1]
  data_ids[i]        = data_ids[j]
  data_weights[i]    = data_weights[j]

  data_levels[j]     = data_levels[k]
  data_points[2*j]   = data_points[2*k]
  data_points[2*j+1] = data_points[2*k+1]
  data_ids[j]        = data_ids[k]
  data_weights[j]    = data_weights[k]

  data_levels[k]     = a_level
  data_points[2*k]   = a_x
  data_points[2*k+1] = a_y
  data_ids[k]        = a_id
  data_weights[k]    = a_weight
}

function shufflePivot(
  i, j,
  a_level, a_x, a_y, a_id, a_weight,
  data_levels, data_points, data_ids, data_weights) {

  data_levels[i]     = data_levels[j]
  data_points[2*i]   = data_points[2*j]
  data_points[2*i+1] = data_points[2*j+1]
  data_ids[i]        = data_ids[j]
  data_weights[i]    = data_weights[j]

  data_levels[j]     = a_level
  data_points[2*j]   = a_x
  data_points[2*j+1] = a_y
  data_ids[j]        = a_id
  data_weights[j]    = a_weight
}

function compare(i, j, data_levels, data_points, data_ids) {
  return ((data_levels[i] - data_levels[j]) ||
          (data_points[2*j] - data_points[2*i]) ||
          (data_ids[i] - data_ids[j])) < 0
}

function comparePivot(i, level, x, y, id, data_levels, data_points, data_ids) {
  return ((level - data_levels[i]) ||
          (data_points[2*i] - x) ||
          (id - data_ids[i])) < 0
}

function quickSort(left, right, data_levels, data_points, data_ids, data_weights) {
  var sixth = (right - left + 1) / 6 | 0,
      index1 = left + sixth,
      index5 = right - sixth,
      index3 = left + right >> 1,
      index2 = index3 - sixth,
      index4 = index3 + sixth,
      el1 = index1,
      el2 = index2,
      el3 = index3,
      el4 = index4,
      el5 = index5,
      less = left + 1,
      great = right - 1,
      tmp = 0
  if(compare(el1, el2, data_levels, data_points, data_ids, data_weights)) {
    tmp = el1
    el1 = el2
    el2 = tmp
  }
  if(compare(el4, el5, data_levels, data_points, data_ids, data_weights)) {
    tmp = el4
    el4 = el5
    el5 = tmp
  }
  if(compare(el1, el3, data_levels, data_points, data_ids, data_weights)) {
    tmp = el1
    el1 = el3
    el3 = tmp
  }
  if(compare(el2, el3, data_levels, data_points, data_ids, data_weights)) {
    tmp = el2
    el2 = el3
    el3 = tmp
  }
  if(compare(el1, el4, data_levels, data_points, data_ids, data_weights)) {
    tmp = el1
    el1 = el4
    el4 = tmp
  }
  if(compare(el3, el4, data_levels, data_points, data_ids, data_weights)) {
    tmp = el3
    el3 = el4
    el4 = tmp
  }
  if(compare(el2, el5, data_levels, data_points, data_ids, data_weights)) {
    tmp = el2
    el2 = el5
    el5 = tmp
  }
  if(compare(el2, el3, data_levels, data_points, data_ids, data_weights)) {
    tmp = el2
    el2 = el3
    el3 = tmp
  }
  if(compare(el4, el5, data_levels, data_points, data_ids, data_weights)) {
    tmp = el4
    el4 = el5
    el5 = tmp
  }

  var pivot1_level  = data_levels[el2]
  var pivot1_x      = data_points[2*el2]
  var pivot1_y      = data_points[2*el2+1]
  var pivot1_id     = data_ids[el2]
  var pivot1_weight = data_weights[el2]

  var pivot2_level  = data_levels[el4]
  var pivot2_x      = data_points[2*el4]
  var pivot2_y      = data_points[2*el4+1]
  var pivot2_id     = data_ids[el4]
  var pivot2_weight = data_weights[el4]

  var ptr0 = el1
  var ptr2 = el3
  var ptr4 = el5
  var ptr5 = index1
  var ptr6 = index3
  var ptr7 = index5

  var level_x = data_levels[ptr0]
  var level_y = data_levels[ptr2]
  var level_z = data_levels[ptr4]
  data_levels[ptr5] = level_x
  data_levels[ptr6] = level_y
  data_levels[ptr7] = level_z

  for (var i1 = 0; i1 < 2; ++i1) {
    var x = data_points[2*ptr0+i1]
    var y = data_points[2*ptr2+i1]
    var z = data_points[2*ptr4+i1]
    data_points[2*ptr5+i1] = x
    data_points[2*ptr6+i1] = y
    data_points[2*ptr7+i1] = z
  }

  var id_x = data_ids[ptr0]
  var id_y = data_ids[ptr2]
  var id_z = data_ids[ptr4]
  data_ids[ptr5] = id_x
  data_ids[ptr6] = id_y
  data_ids[ptr7] = id_z

  var weight_x = data_weights[ptr0]
  var weight_y = data_weights[ptr2]
  var weight_z = data_weights[ptr4]
  data_weights[ptr5] = weight_x
  data_weights[ptr6] = weight_y
  data_weights[ptr7] = weight_z

  move(index2, left,  data_levels, data_points, data_ids, data_weights)
  move(index4, right, data_levels, data_points, data_ids, data_weights)
  for (var k = less; k <= great; ++k) {
    if (comparePivot(k,
        pivot1_level, pivot1_x, pivot1_y, pivot1_id,
        data_levels, data_points, data_ids)) {
      if (k !== less) {
        swap(k, less, data_levels, data_points, data_ids, data_weights)
      }
      ++less;
    } else {
      if (!comparePivot(k,
          pivot2_level, pivot2_x, pivot2_y, pivot2_id,
          data_levels, data_points, data_ids)) {
        while (true) {
          if (!comparePivot(great,
              pivot2_level, pivot2_x, pivot2_y, pivot2_id,
              data_levels, data_points, data_ids)) {
            if (--great < k) {
              break;
            }
            continue;
          } else {
            if (comparePivot(great,
                pivot1_level, pivot1_x, pivot1_y, pivot1_id,
                data_levels, data_points, data_ids)) {
              rotate(k, less, great, data_levels, data_points, data_ids, data_weights)
              ++less;
              --great;
            } else {
              swap(k, great, data_levels, data_points, data_ids, data_weights)
              --great;
            }
            break;
          }
        }
      }
    }
  }
  shufflePivot(left, less-1, pivot1_level, pivot1_x, pivot1_y, pivot1_id, pivot1_weight, data_levels, data_points, data_ids, data_weights)
  shufflePivot(right, great+1, pivot2_level, pivot2_x, pivot2_y, pivot2_id, pivot2_weight, data_levels, data_points, data_ids, data_weights)
  if (less - 2 - left <= INSERT_SORT_CUTOFF) {
    insertionSort(left, less - 2, data_levels, data_points, data_ids, data_weights)
  } else {
    quickSort(left, less - 2, data_levels, data_points, data_ids, data_weights)
  }
  if (right - (great + 2) <= INSERT_SORT_CUTOFF) {
    insertionSort(great + 2, right, data_levels, data_points, data_ids, data_weights)
  } else {
    quickSort(great + 2, right, data_levels, data_points, data_ids, data_weights)
  }
  if (great - less <= INSERT_SORT_CUTOFF) {
    insertionSort(less, great, data_levels, data_points, data_ids, data_weights)
  } else {
    quickSort(less, great, data_levels, data_points, data_ids, data_weights)
  }
}
'use strict'

var _$normalize_19 = __normalize_19;

function __normalize_19 (arr, dim) {
	if (!arr || arr.length == null) { throw Error('Argument should be an array') }

	if (dim == null) { dim = 1 }
	else { dim = Math.floor(dim) }

	var bounds = Array(dim * 2)

	for (var offset = 0; offset < dim; offset++) {
		var max = -Infinity, min = Infinity, i = offset, l = arr.length;

		for (; i < l; i+=dim) {
			if (arr[i] > max) { max = arr[i]; }
			if (arr[i] < min) { min = arr[i]; }
		}

		bounds[offset] = min
		bounds[dim + offset] = max
	}

	return bounds;
}
'use strict'

/* removed: var _$sortLevels_18 = require('./lib/sort') */;
/* removed: var _$normalize_19 = require('array-bounds') */;

var _$snapPoints_20 = snapPoints

function partition(points, ids, start, end, lox, loy, hix, hiy) {
  var mid = start
  for(var i=start; i<end; ++i) {
    var x  = points[2*i]
    var y  = points[2*i+1]
    var s  = ids[i]
    if(lox <= x && x <= hix &&
       loy <= y && y <= hiy) {
      if(i === mid) {
        mid += 1
      } else {
        points[2*i]     = points[2*mid]
        points[2*i+1]   = points[2*mid+1]
        ids[i]          = ids[mid]
        points[2*mid]   = x
        points[2*mid+1] = y
        ids[mid]        = s
        mid += 1
      }
    }
  }
  return mid
}

function SnapInterval(pixelSize, offset, count) {
  this.pixelSize  = pixelSize
  this.offset     = offset
  this.count      = count
}

function snapPoints(points, ids, weights, bounds) {
  var n = points.length >>> 1
  if(n < 1) {
    return []
  }

  if (!ids) { ids = Array(n) }
  if (!weights) { weights = Array(n) }
  if (!bounds) { bounds = [] }

  for(var i=0; i<n; ++i) {
    ids[i] = i
  }

  // empty bounds or invalid bounds are considered as undefined and require recalc
  if (!bounds.length || bounds.length < 4 || bounds[0] >= bounds[2] || bounds[1] >= bounds[3]) {
    var b = _$normalize_19(points, 2)

    if(b[0] === b[2]) {
      b[2] += 1
    }
    if(b[1] === b[3]) {
      b[3] += 1
    }

    bounds[0] = b[0]
    bounds[1] = b[1]
    bounds[2] = b[2]
    bounds[3] = b[3]
  }

  var lox = bounds[0]
  var loy = bounds[1]
  var hix = bounds[2]
  var hiy = bounds[3]

  //Calculate diameter
  var scaleX = 1.0 / (hix - lox)
  var scaleY = 1.0 / (hiy - loy)
  var diam = Math.max(hix - lox, hiy - loy)



  var levels = new Int32Array(n)
  var ptr = 0

  function snapRec(x, y, diam, start, end, level) {
    var diam_2 = diam * 0.5
    var offset = start + 1
    var count = end - start
    weights[ptr] = count
    levels[ptr++] = level
    for(var i=0; i<2; ++i) {
      for(var j=0; j<2; ++j) {
        var nx = x+i*diam_2
        var ny = y+j*diam_2
        var nextOffset = partition(
            points
          , ids
          , offset
          , end
          , nx, ny
          , nx+diam_2, ny+diam_2)
        if(nextOffset === offset) {
          continue
        }
        if(nextOffset - offset >= Math.max(0.9 * count, 32)) {
          var mid = (end + start)>>>1
          snapRec(nx, ny, diam_2, offset, mid, level+1)
          offset = mid
        }
        snapRec(nx, ny, diam_2, offset, nextOffset, level+1)
        offset = nextOffset
      }
    }
  }
  snapRec(lox, loy, diam, 0, n, 0)
  _$sortLevels_18(levels, points, ids, weights, n)

  var lod         = []
  var lastLevel   = 0
  var prevOffset  = n
  for(var ptr=n-1; ptr>=0; --ptr) {
    points[2*ptr]   = (points[2*ptr]   - lox) * scaleX
    points[2*ptr+1] = (points[2*ptr+1] - loy) * scaleY

    var level = levels[ptr]
    if(level === lastLevel) {
      continue
    }

    lod.push(new SnapInterval(
      diam * Math.pow(0.5, level),
      ptr+1,
      prevOffset - (ptr+1)
    ))
    prevOffset = ptr+1

    lastLevel = level
  }

  lod.push(new SnapInterval(diam * Math.pow(0.5, level+1), 0, prevOffset))

  return lod
}
'use strict'

var _$updateDiff_21 = function updateDiff (obj, diff, mappers) {
	if (!Array.isArray(mappers)) { mappers = [].slice.call(arguments, 2) }

	for (var i = 0, l = mappers.length; i < l; i++) {
		var dict = mappers[i]
		for (var prop in dict) {
			if (diff[prop] !== undefined && !Array.isArray(diff[prop]) && obj[prop] === diff[prop]) { continue }

			if (prop in diff) {
				var result

				if (dict[prop] === true) { result = diff[prop] }
				else if (dict[prop] === false) { continue }
				else if (typeof dict[prop] === 'function') {
					result = dict[prop](diff[prop], obj, diff)
					if (result === undefined) { continue }
				}

				obj[prop] = result
			}
		}
	}

	return obj
}
'use strict'

/* removed: var _$normalize_6 = require('color-normalize') */;
/* removed: var _$normalize_1 = require('array-bounds') */;
/* removed: var _$module_4 = require('color-id') */;
/* removed: var _$snapPoints_20 = require('snap-points-2d') */;
/* removed: var _$module_16 = require('object-assign') */;
/* removed: var _$module_14 = require('glslify') */;
/* removed: var _$module_2 = require('binary-search-bounds') */;
/* removed: var _$pick_17 = require('pick-by-alias') */;
/* removed: var _$updateDiff_21 = require('update-diff') */;
/* removed: var _$flattenVertexData_13 = require('flatten-vertex-data') */;

var _$Scatter_IndexJs = Scatter


function Scatter (regl, options) {
	if (typeof regl === 'function') {
		if (!options) { options = {} }
		options.regl = regl
	}
	else {
		options = regl
		regl = null
	}

	if (options && options.length) { options.positions = options }
	regl = options.regl

	var hasElementIndex = regl.hasExtension('OES_element_index_uint')

	// persistent variables
	var gl = regl._gl,
		drawMarker, drawCircle,
		sizeBuffer, positionBuffer, positionFractBuffer, colorBuffer,
		paletteTexture, palette = [], paletteIds = {},
		defaults = {
			color: 'black',
			borderColor: 'transparent',
			borderSize: 1,
			size: 12,
			opacity: 1,
			marker: undefined,
			viewport: null,
			range: null,
			pixelSize: null,
			offset: 0,
			count: 0,
			bounds: null,
			positions: [],
			snap: 1e4
		},
		groups = [],

		//textures for marker keys
		markerTextures = [null],
		markerCache = [null]

	var maxColors = 256, maxSize = 100

	//texture with color palette
	paletteTexture = regl.texture({
		width: maxColors,
		height: 1,
		type: 'uint8',
		format: 'rgba',
		mag: 'nearest',
		min: 'nearest'
	})

	//buffers to reuse
	sizeBuffer = regl.buffer({
		usage: 'dynamic',
		type: 'float',
		data: null
	})
	colorBuffer = regl.buffer({
		usage: 'dynamic',
		type: 'uint8',
		data: null
	})
	positionBuffer = regl.buffer({
		usage: 'dynamic',
		type: 'float',
		data: null
	})
	positionFractBuffer = regl.buffer({
		usage: 'dynamic',
		type: 'float',
		data: null
	})

	//init defaults
	update(options)


	//common shader options
	var shaderOptions = {
		uniforms: {
			pixelRatio: regl.context('pixelRatio'),
			palette: paletteTexture,
			scale: regl.prop('scale'),
			scaleFract: regl.prop('scaleFract'),
			translate: regl.prop('translate'),
			translateFract: regl.prop('translateFract'),
			opacity: regl.prop('opacity'),
			marker: regl.prop('marker')
		},

		attributes: {
			position: positionBuffer,
			positionFract: positionFractBuffer,
			size: function (ctx, prop) { return prop.size.length ? {
				buffer: sizeBuffer,
				stride: 2,
				offset: 0
			} : {constant: [Math.round(prop.size * 255 / maxSize)]}; },
			borderSize: function (ctx, prop) { return prop.borderSize.length ? {
				buffer: sizeBuffer,
				stride: 2,
				offset: 1
			} : {constant: [Math.round(prop.borderSize * 255 / maxSize)]}; },
			colorId: function (ctx, prop) { return prop.color.length ? {
				buffer: colorBuffer,
				stride: 2,
				offset: 0
			} : {constant: [prop.color]}; },
			borderColorId: function (ctx, prop) { return prop.borderColor.length ? {
				buffer: colorBuffer,
				stride: 2,
				offset: 1
			} : {constant: [prop.borderColor]}; }
		},

		blend: {
			enable: true,
			color: [0,0,0,1],
			func: {
				srcRGB: 'src alpha',
				dstRGB: 'one minus src alpha',
				srcAlpha: 'one minus dst alpha',
				dstAlpha: 'one'
			}
		},

		scissor: {
			enable: true,
			box: regl.prop('viewport')
		},
		stencil: false,
		viewport: regl.prop('viewport'),

		depth: {
			enable: false
		},

		elements: regl.prop('elements'),
		count: regl.prop('count'),
		offset: regl.prop('offset'),

		primitive: 'points'
	}

	//draw sdf-marker
	var markerOptions = _$module_16({}, shaderOptions)
	markerOptions.frag = _$module_14(["precision mediump float;\n#define GLSLIFY 1\n\nvarying vec4 fragColor, fragBorderColor;\nvarying float fragWidth, fragBorderColorLevel, fragColorLevel;\n\nuniform sampler2D marker;\nuniform float pixelRatio, opacity;\n\nfloat smoothStep(float x, float y) {\n  return 1.0 / (1.0 + exp(50.0*(x - y)));\n}\n\nvoid main() {\n  float dist = texture2D(marker, gl_PointCoord).r, delta = fragWidth;\n\n  //max-distance alpha\n  if (dist < 0.003) discard;\n\n  //null-border case\n  if (fragBorderColorLevel == fragColorLevel || fragBorderColor.a == 0.) {\n    float colorAmt = smoothstep(.5 - delta, .5 + delta, dist);\n    gl_FragColor = vec4(fragColor.rgb, colorAmt * fragColor.a);\n    return;\n  }\n\n  float borderColorAmt = smoothstep(fragBorderColorLevel - delta, fragBorderColorLevel + delta, dist);\n  float colorAmt = smoothstep(fragColorLevel - delta, fragColorLevel + delta, dist);\n\n  vec4 color = fragBorderColor;\n  color.a *= borderColorAmt;\n  color = mix(color, fragColor, colorAmt);\n  color.a *= opacity;\n\n  gl_FragColor = color;\n}\n"])
	markerOptions.vert = _$module_14(["precision mediump float;\n#define GLSLIFY 1\n\nattribute vec2 position, positionFract;\nattribute float size, borderSize;\nattribute float colorId, borderColorId;\n\nuniform vec2 scale, scaleFract, translate, translateFract;\nuniform float pixelRatio;\nuniform sampler2D palette;\n\nconst float paletteSize = 256., maxSize = 100.;\nconst float borderLevel = .5;\n\nvarying vec4 fragColor, fragBorderColor;\nvarying float fragPointSize, fragBorderRadius,\n\t\tfragWidth, fragBorderColorLevel, fragColorLevel;\n\nvoid main() {\n  vec4 color = texture2D(palette, vec2((colorId + .5) / paletteSize, 0));\n  vec4 borderColor = texture2D(palette, vec2((borderColorId + .5) / paletteSize, 0));\n\n  float size = size * maxSize / 255.;\n  float borderSize = borderSize * maxSize / 255.;\n\n  gl_PointSize = 2. * size * pixelRatio;\n  fragPointSize = size * pixelRatio;\n\n  vec2 pos = (position + translate) * scale\n      + (positionFract + translateFract) * scale\n      + (position + translate) * scaleFract\n      + (positionFract + translateFract) * scaleFract;\n\n  gl_Position = vec4(pos * 2. - 1., 0, 1);\n\n  fragColor = color;\n  fragBorderColor = borderColor;\n  fragWidth = 1. / gl_PointSize;\n\n  fragBorderColorLevel = clamp(borderLevel - borderLevel * borderSize / size, 0., 1.);\n  fragColorLevel = clamp(borderLevel + (1. - borderLevel) * borderSize / size, 0., 1.);\n}\n"])

	drawMarker = regl(markerOptions)

	//draw circle
	var circleOptions = _$module_16({}, shaderOptions)
	circleOptions.frag = _$module_14(["precision highp float;\n#define GLSLIFY 1\n\nvarying vec4 fragColor, fragBorderColor;\nvarying float fragBorderRadius, fragWidth;\n\nuniform float pixelRatio, opacity;\n\nvoid main() {\n\tfloat radius, alpha = 1.0, delta = fragWidth;\n\n\tradius = length(2.0 * gl_PointCoord - 1.0);\n\n\tif(radius > 1.0 + delta) {\n\t\tdiscard;\n\t\treturn;\n\t}\n\n\talpha -= smoothstep(1.0 - delta, 1.0 + delta, radius);\n\n\tvec4 color = mix(fragColor, fragBorderColor, smoothstep(fragBorderRadius - delta, fragBorderRadius + delta, radius));\n\tcolor.a *= alpha * opacity;\n\tgl_FragColor = color;\n}\n"])
	circleOptions.vert = _$module_14(["precision highp float;\n#define GLSLIFY 1\n\nattribute vec2 position, positionFract;\nattribute float size, borderSize;\nattribute float colorId, borderColorId;\n\nuniform vec2 scale, scaleFract, translate, translateFract;\nuniform float pixelRatio;\nuniform sampler2D palette;\n\nconst float paletteSize = 256., maxSize = 100.;\n\nvarying vec4 fragColor, fragBorderColor;\nvarying float fragBorderRadius, fragWidth;\n\nvoid main() {\n  vec4 color = texture2D(palette, vec2((colorId + .5) / paletteSize, 0));\n  vec4 borderColor = texture2D(palette, vec2((borderColorId + .5) / paletteSize, 0));\n\n  float size = size * maxSize / 255.;\n  float borderSize = borderSize * maxSize / 255.;\n\n  gl_PointSize = (size + borderSize) * pixelRatio;\n\n  vec2 pos = (position + translate) * scale\n\t\t\t+ (positionFract + translateFract) * scale\n\t\t\t+ (position + translate) * scaleFract\n\t\t\t+ (positionFract + translateFract) * scaleFract;\n\n  gl_Position = vec4(pos * 2. - 1., 0, 1);\n\n  fragBorderRadius = borderSize == 0. ? 2. : 1. - 2. * borderSize / (size + borderSize);\n  fragColor = color;\n  fragBorderColor = borderColor;\n  fragWidth = 1. / gl_PointSize;\n}\n"])

	drawCircle = regl(circleOptions)

	//expose API
	_$module_16(scatter2d, {
		update: update,
		draw: draw,
		destroy: destroy,
		regl: regl,
		gl: gl,
		canvas: gl.canvas,
		groups: groups
	})


	function scatter2d (opts) {
		//update
		if (opts) {
			update(opts)
		}

		//destroy
		else if (opts === null) {
			destroy()
		}

		draw(opts)
	}

	function draw (opts) {
		if (typeof opts === 'number') { return drawGroup(opts) }
		if (Array.isArray(opts[0])) {
			opts.forEach(function (els, i) {
				if (els == null) { return }
				if (els.length) { return drawGroup(els, i) }
				return drawGroup(els)
			})
			return
		}

		//make options a batch
		groups.forEach(function (group, i) {
			if (!group) { return }

			drawGroup(i)
		})
	}

	function drawGroup (group, id) {
		if (typeof group === 'number') { group = groups[group] }

		var els
		if (Array.isArray(group)) {
			els = group
			group = groups[id]
		}

		if (!(group && group.count && group.opacity)) { return }

		//draw subset of elements
		if (els) {
			var pending = {};

			for (var i = 0; i < els.length; i++) {
				pending[els[i]] = true
			}

			var batch$1 = [], offset = group.offset
			for (var i$1 = 0; i$1 < group.markerIds.length; i$1++) {
				var subIds = [], ids = group.markerIds[i$1]

				for (var i$2 = 0, l = ids.length; i$2 < l; i$2++) {
					var id$1 = ids[i$2]
					if (pending[id$1]) {
						subIds.push(id$1 + offset)
						pending[id$1] = null
					}
				}
				batch$1.push(_$module_16({}, group, {
					elements: subIds,
					offset: 0,
					count: subIds.length,
					marker: markerTextures[ids.id]
				}))
			}

			regl._refresh()
			drawCircle(batch$1.shift())
			regl._refresh()
			drawMarker(batch$1)

			return
		}


		//draw circles
		//FIXME remove regl._refresh hooks once regl issue #427 is fixed
		if (group.markerIds[0]) {
			regl._refresh()
			drawCircle(getMarkerDrawOptions(group.markerIds[0], group))
		}

		//draw all other available markers
		var batch = []
		for (var i$3 = 1; i$3 < group.markerIds.length; i$3++) {
			var ids$1 = group.markerIds[i$3]

			if (!ids$1 || !ids$1.length) { continue }

			[].push.apply(batch, getMarkerDrawOptions(ids$1, group))
		}

		if (batch.length) {
			regl._refresh()
			drawMarker(batch)
		}
	}

	//get options for the marker ids
	function getMarkerDrawOptions(ids, group) {
		//unsnapped options
		if (!ids.snap) {
			return [_$module_16({}, group, {
				elements: ids.elements,
				offset: 0,
				count: ids.length,
				marker: markerTextures[ids.id]
			})]
		}

		//scales batch
		var batch = []
		var range = group.range;
		var lod = ids.lod;
		var x = ids.x;
		var id = ids.id;
		var pixelSize = (range[2] - range[0]) / group.viewport.width

		var els = ids.elements

		for (var scaleNum = lod.length; scaleNum--;) {
			var level = lod[scaleNum]

			//FIXME: use minSize-adaptive coeff here, if makes sense, mb we need dist tho
			if (level.pixelSize && level.pixelSize < pixelSize && scaleNum > 1) {
				continue
			}

			var intervalStart = level.offset
			var intervalEnd = level.count + intervalStart

			var startOffset = _$module_2.ge(x, range[0], intervalStart, intervalEnd - 1)
			var endOffset = _$module_2.lt(x, range[2], startOffset, intervalEnd - 1) + 1

			if (endOffset <= startOffset) { continue }

			batch.push(_$module_16({}, group, {
				elements: els,
				marker: markerTextures[id],
				offset: startOffset,
				count: endOffset - startOffset
			}))
		}

		return batch
	}

	function update (options) {
		if (!options) { return }

		//direct points argument
		if (options.length != null) {
			if (typeof options[0] === 'number') { options = [{positions: options}] }
		}
		//make options a batch
		else if (!Array.isArray(options)) { options = [options] }

		//global count of points
		var pointCount = 0, sizeCount = 0, colorCount = 0

		groups = options.map(function (options, i) {
			var group = groups[i]

			if (!options) { return group }
			else if (typeof options === 'function') { options = {after: options} }
			else if (typeof options[0] === 'number') { options = {positions: options} }

			//copy options to avoid mutation & handle aliases
			options = _$pick_17(options, {
				positions: 'positions data points',
				snap: 'snap cluster',
				size: 'sizes size radius',
				borderSize: 'borderSizes borderSize stroke-width strokeWidth outline',
				color: 'colors color fill fill-color fillColor',
				borderColor: 'borderColors borderColor stroke stroke-color strokeColor',
				palette: 'palette swatch',
				marker: 'markers marker shape',
				range: 'range dataBox',
				viewport: 'viewport viewBox',
				opacity: 'opacity alpha'
			})
			if (!group) {
				groups[i] = group = {
					id: i,
					scale: null,
					translate: null,
					scaleFract: null,
					translateFract: null,

					//list of ids corresponding to markers, with inner props
					markerIds: []
				}
				options = _$module_16({}, defaults, options)
			}

			//force update triggers
			if (options.positions && !('marker' in options)) {
				options.marker = group.marker
				delete group.marker
			}

			_$updateDiff_21(group, options, [{
				snap: true,
				size: function (s) {
					sizeCount += s.length ? 1 : 0
					return s
				},
				borderSize: function (s) {
					sizeCount += s.length ? 1 : 0
					return s
				},
				opacity: parseFloat,

				//add colors to palette, save references
				color: function (c) {
					c = updateColor(c)
					colorCount++
					return c
				},
				borderColor: function (c) {
					c = updateColor(c)
					colorCount++
					return c
				},

				positions: function (positions, group, options) {
					positions = _$flattenVertexData_13(positions, 'float64')

					var count = group.count = Math.floor(positions.length / 2)
					var bounds = group.bounds = count ? _$normalize_1(positions, 2) : null

					// if range is not provided updated - recalc it
					if (!options.range && !group.range) {
						delete group.range
						options.range = bounds
					}

					group.offset = pointCount
					pointCount += count

					return positions
				}
			}, {
				//create marker ids corresponding to known marker textures
				marker: function (markers, group, options) {
					//reset marker elements
					group.markerIds.length = 0

					//single sdf marker
					if (!markers || typeof markers[0] === 'number') {
						var id = addMarker(markers)

						var elements = Array(group.count)
						for (var i = 0; i < group.count; i++) {
							elements[i] = i
						}

						group.markerIds[id] = elements
					}
					//per-point markers
					else {
						for (var i$1 = 0, l = markers.length; i$1 < l; i$1++) {
							var id$1 = addMarker(markers[i$1])

							if (!group.markerIds[id$1]) { group.markerIds[id$1] = [] }
							group.markerIds[id$1].push(i$1)
						}
					}

					return markers
				}
			}, {
				//recalculate per-marker snapping
				//first, it is faster to snap 100 points 100 times than 10000 points once (practically, not theoretically)
				//second, it is easier to subset render per-marker than per-generic set
				positions: function (positions, group) {
					var markerIds = group.markerIds;
					var snap = group.snap;
					var bounds = group.bounds;
					var offset = group.offset;

					for (var i = 0; i < markerIds.length; i++) {
						var ids = markerIds[i]
						if (!ids || !ids.length) { continue }

						var l = ids.length, els = (void 0)

						ids.id = i;

						if (snap && (snap === true || l * 2 > snap)) {
							ids.snap = true
							var x = ids.x = Array(l)
							var w = ids.w = Array(l)
							var markerPoints = (void 0)

							//multimarker snapping is computationally more intense
							if (markerIds.length > 1) {
								markerPoints = Array(l * 2)

								for (var i$1 = 0; i$1 < l; i$1++) {
									var id = ids[i$1]
									markerPoints[i$1 * 2] = positions[id * 2]
									markerPoints[i$1 * 2 + 1] = positions[id * 2 + 1]
								}
							}
							else {
								markerPoints = new Float64Array(positions.length)
								markerPoints.set(positions)
							}

							//shuffled_id: real_id
							var i2id = new Uint32Array(l)

							ids.lod = _$snapPoints_20(markerPoints, i2id, w, bounds)

							els = new Uint32Array(l)
							for (var i$2 = 0; i$2 < l; i$2++) {
								var id$1 = i2id[i$2], iid = ids[id$1]
								els[i$2] = iid + offset
								x[i$2] = positions[iid * 2]
							}
						}
						else {
							els = new Uint32Array(l)
							for (var i$3 = 0; i$3 < l; i$3++) {
								els[i$3] = ids[i$3] + offset
							}
						}

						ids.elements = regl.elements({
							primitive: 'points',
							type: hasElementIndex ? 'uint32' : 'uint16',
							data: els
						})
					}
				},

				range: function (range, group, options) {
					var bounds = group.bounds

					if (!bounds) { return }
					if (!range) { range = bounds }

					group.scale = [1 / (range[2] - range[0]), 1 / (range[3] - range[1])]
					group.translate = [-range[0], -range[1]]

					group.scaleFract = fract32(group.scale)
					group.translateFract = fract32(group.translate)

					return range
				},

				viewport: function (vp) {
					var viewport

					if (Array.isArray(vp)) {
						viewport = {
							x: vp[0],
							y: vp[1],
							width: vp[2] - vp[0],
							height: vp[3] - vp[1]
						}
					}
					else if (vp) {
						viewport = {
							x: vp.x || vp.left || 0,
							y: vp.y || vp.top || 0
						}

						if (vp.right) { viewport.width = vp.right - viewport.x }
						else { viewport.width = vp.w || vp.width || 0 }

						if (vp.bottom) { viewport.height = vp.bottom - viewport.y }
						else { viewport.height = vp.h || vp.height || 0 }
					}
					else {
						viewport = {
							x: 0, y: 0,
							width: gl.drawingBufferWidth,
							height: gl.drawingBufferHeight
						}
					}

					return viewport
				}
			}])

			return group
		})

		//put point/color data into buffers, if updated any of them
		var len = groups.reduce(function (acc, group, i) {
			return acc + (group ? group.count : 0)
		}, 0)

		if (pointCount) {
			var positionData = new Float32Array(len * 2)
			var positionFractData = new Float32Array(len * 2)

			groups.forEach(function (group, i) {
				if (!group) { return }
				var positions = group.positions;
				var count = group.count;
				var offset = group.offset;
				if (!count) { return }
				positionData.set(float32(positions), offset * 2)
				positionFractData.set(fract32(positions), offset * 2)
			})

			positionBuffer(positionData)
			positionFractBuffer(positionFractData)
		}

		if (sizeCount) {
			var sizeData = new Uint8Array(len * 2)

			groups.forEach(function (group, i) {
				if (!group) { return }
				var count = group.count;
				var offset = group.offset;
				var size = group.size;
				var borderSize = group.borderSize;
				if (!count) { return }

				if (size.length || borderSize.length) {
					var sizes = new Uint8Array(count*2)
					for (var i$1 = 0; i$1 < count; i$1++) {
						//we downscale size to allow for fractions
						sizes[i$1*2] = Math.round((size[i$1] == null ? size : size[i$1]) * 255 / maxSize)
						sizes[i$1*2 + 1] = Math.round((borderSize[i$1] == null ? borderSize : borderSize[i$1]) * 255 / maxSize)
					}
					sizeData.set(sizes, offset * 2)
				}
			})
			sizeBuffer(sizeData)
		}

		if (colorCount) {
			var colorData = new Uint8Array(len * 2)

			groups.forEach(function (group, i) {
				if (!group) { return }
				var count = group.count;
				var offset = group.offset;
				var color = group.color;
				var borderColor = group.borderColor;
				if (!count) { return }

				if (color.length || borderColor.length) {
					var colorIds = new Uint8Array(count*2)
					for (var i$1 = 0; i$1 < count; i$1++) {
						colorIds[i$1*2] = color[i$1] == null ? color : color[i$1]
						colorIds[i$1*2 + 1] = borderColor[i$1] == null ? borderColor : borderColor[i$1]
					}
					colorData.set(colorIds, offset * 2)
				}
			})

			colorBuffer(colorData)
		}
	}

	//get (and create) marker texture id
	function addMarker (sdf) {
		var pos = sdf == null ? 0 : markerCache.indexOf(sdf)

		if (pos >= 0) { return pos }

		//convert sdf to 0..255 range
		var distArr
		if (sdf instanceof Uint8Array || sdf instanceof Uint8ClampedArray) {
			distArr = sdf
		}
		else {
			distArr = new Uint8Array(sdf.length)
			for (var i = 0, l = sdf.length; i < l; i++) {
				distArr[i] = sdf[i] * 255
			}
		}

		var radius = Math.floor(Math.sqrt(distArr.length))

		pos = markerTextures.length

		markerCache.push(sdf)
		markerTextures.push(regl.texture({
			channels: 1,
			data: distArr,
			radius: radius,
			mag: 'linear',
			min: 'linear'
		}))

		return pos
	}

	//register color to palette, return it's index or list of indexes
	function updateColor (colors) {
		if (!Array.isArray(colors)) {
			colors = [colors]
		}

		var start = palette.length
		var idx = []
		for (var i = 0; i < colors.length; i++) {
			var color = colors[i]

			//idx colors
			if (typeof color === 'number') {
				idx[i] = color
				continue
			}

			color = _$normalize_6(color, 'uint8')

			var id = _$module_4(color, false)
			if (paletteIds[id] == null) {
				var pos = palette.length
				paletteIds[id] = Math.floor(pos / 4)
				palette[pos] = color[0]
				palette[pos+1] = color[1]
				palette[pos+2] = color[2]
				palette[pos+3] = color[3]
			}
			idx[i] = paletteIds[id]
		}

		var w = (palette.length - start) * .25

		if (w > 0) {
			paletteTexture.subimage({
				width: w,
				height: 1,
				data: palette.slice(start)
			}, start * .25, 0)
		}

		//keep static index for single-color property
		return idx.length === 1 ? idx[0] : idx
	}

	function destroy () {
		sizeBuffer.destroy()
		positionBuffer.destroy()
		positionFractBuffer.destroy()
		colorBuffer.destroy()
		paletteTexture.destroy()
	}

	return scatter2d
}

//return fractional part of float32 array
function fract32 (arr) {
	var fract = new Float32Array(arr.length)
	fract.set(arr)
	for (var i = 0, l = fract.length; i < l; i++) {
		fract[i] = arr[i] - fract[i]
	}
	return fract
}
function float32 (arr) {
	if (arr instanceof Float32Array) { return arr }

	var float = new Float32Array(arr)
	float.set(arr)
	return float
}
}());
